"""Job to execute"""

from __future__ import annotations

import shlex
from typing import Any, Tuple

from .defaults import JobStatus
from .utils import logger, CommandType
from .path import SpecPath


class Job:
    """The class for job

    Attributes:

        cmd: The command
        index: The index of the job
        metadir: The metadir of the job
        jid: The jid of the job in scheduler system
        trial_count: The count for re-tries
        _status: The status of the job
        _rc: The return code of the job
        _error_retry: Whether we should retry if error happened
        _num_retries: Total number of retries

    Args:
        index: The index of the job
        cmd: The command of the job
        metadir: The meta directory of the Job
        error_retry: Whether we should retry if error happened
        num_retries: Total number of retries
    """

    __slots__ = (
        "cmd",
        "index",
        "metadir",
        "trial_count",
        "_jid",
        "_status",
        "_rc",
        "_error_retry",
        "_num_retries",
        "envs",
    )

    def __init__(
        self,
        index: int,
        cmd: CommandType,
        workdir: SpecPath,
        error_retry: bool | None = None,
        num_retries: int | None = None,
        envs: dict[str, Any] | None = None,
    ):
        """Construct a new Job

        Args:
            index: The index of the job
            cmd: The command of the job
            metadir: The meta directory of the Job
            error_retry: Whether we should retry if error happened
            num_retries: Total number of retries
        """
        self.cmd: Tuple[str, ...] = tuple(
            map(
                str,
                (cmd if isinstance(cmd, (tuple, list)) else shlex.split(cmd)),
            )
        )
        self.index = index
        self.envs = envs or {}
        self.envs["XQUTE_JOB_INDEX"] = str(self.index)
        self.envs["XQUTE_METADIR"] = str(workdir)
        self.metadir = workdir / str(self.index)  # type: ignore
        self.envs["XQUTE_JOB_METADIR"] = str(self.metadir)
        # For cloud paths, this requires cloud client
        # self.metadir.mkdir(exist_ok=True, parents=True)
        # Let Scheduler.create_job handle metadir creation

        # The name of the job, should be the unique id from the scheduler
        self.trial_count = 0

        self._jid: int | str | None = None
        self._status = JobStatus.INIT
        self._error_retry = error_retry
        self._num_retries = num_retries

    def __repr__(self) -> str:
        """repr of the job"""
        prefix = f"{self.__class__.__name__}-{self.index}"
        if not self._jid:
            return f"<{prefix}: ({self.cmd})>"
        return f"<{prefix}({self._jid}): ({self.cmd})>"

    @property
    def stdout_file(self) -> SpecPath:
        """The stdout file of the job"""
        return self.metadir / "job.stdout"

    @property
    def stderr_file(self) -> SpecPath:
        """The stderr file of the job"""
        return self.metadir / "job.stderr"

    @property
    def status_file(self) -> SpecPath:
        """The status file of the job"""
        return self.metadir / "job.status"

    @property
    def rc_file(self) -> SpecPath:
        """The rc file of the job"""
        return self.metadir / "job.rc"

    @property
    def jid_file(self) -> SpecPath:
        """The jid file of the job"""
        return self.metadir / "job.jid"

    @property
    def retry_dir(self) -> SpecPath:
        """The retry directory of the job"""
        return self.metadir / "job.retry"

    async def get_jid(self) -> int | str | None:
        """Get the jid of the job in scheduler system"""
        if self._jid is None and not await self.jid_file.a_is_file():
            return None
        if self._jid is not None:
            return self._jid
        self._jid = await self.jid_file.a_read_text()
        return self._jid

    async def set_jid(self, uniqid: int | str):
        self._jid = uniqid
        await self.jid_file.a_write_text(str(uniqid))

    async def get_status(self, refresh: bool = False) -> int:
        """Query the status of the job

        If the job is submitted, try to query it from the status file
        Make sure the status is updated by trap in wrapped script

        Uses caching to avoid excessive file I/O. Cache is invalidated
        when status is explicitly set.

        Args:
            refresh: Whether to refresh the status from file
        """
        if not refresh:
            return self._status

        if await self.status_file.a_is_file() and self._status in (
            JobStatus.SUBMITTED,
            JobStatus.RUNNING,
            JobStatus.KILLING,
        ):
            try:
                status_text = await self.status_file.a_read_text()
                self._status = int(status_text)
            except (
                FileNotFoundError,
                ValueError,
                TypeError,
            ):  # pragma: no cover
                pass

        # Don't log here - let scheduler handle transition logging
        return self._status

    async def set_status(self, stat: int, flush: bool = True) -> None:
        """Set the status manually

        Args:
            stat: The status to set
            flush: Whether to flush the status to file
        """
        # Only log if status is actually changing
        prev_status = self._status

        if stat != prev_status:
            logger.info(
                "/Job-%s Status changed: %r -> %r",
                self.index,
                *JobStatus.get_name(prev_status, stat),
            )
            self._status = stat
            if flush:
                await self.status_file.a_write_text(str(stat))

    async def get_rc(self) -> int:
        """The return code of the job"""
        if not await self.rc_file.a_is_file():
            return -9
        return int(await self.rc_file.a_read_text())

    async def set_rc(self, rc: int | str) -> None:
        """Set the return code of the job

        Args:
            rc: The return code
        """
        await self.rc_file.a_write_text(str(rc))

    async def clean(self, retry: bool = False) -> None:
        """Clean up the meta files

        Args:
            retry: Whether clean it for retrying
        """
        files_to_clean = [
            self.stdout_file,
            self.stderr_file,
            self.status_file,
            self.rc_file,
        ]

        if retry:
            retry_dir = self.retry_dir / str(self.trial_count)  # type: ignore
            if await retry_dir.a_exists():
                await retry_dir.a_rmtree()
            await retry_dir.a_mkdir(parents=True)

            for file in files_to_clean:
                if await file.a_is_file():
                    await file.a_rename(retry_dir / file.name)
        else:
            for file in files_to_clean:
                if await file.a_is_file():
                    await file.a_unlink()



"""The xqute module"""

from __future__ import annotations

import asyncio
import functools
import signal
from collections import deque
from typing import TYPE_CHECKING, Any, List, Mapping, Type

from .defaults import (
    DEFAULT_WORKDIR,
    DEFAULT_ERROR_STRATEGY,
    DEFAULT_NUM_RETRIES,
    DEFAULT_SCHEDULER_FORKS,
    # DEFAULT_SUBMISSION_BATCH,
    JobStatus,
    SLEEP_INTERVAL_PRODUCER_MAX_FORKS,
    SLEEP_INTERVAL_POLLING_JOBS,
    SLEEP_INTERVAL_KEEP_FEEDING,
)
from .utils import logger, CommandType
from .plugin import plugin
from .schedulers import get_scheduler

if TYPE_CHECKING:
    from .path import PathType
    from .scheduler import Scheduler
    from .job import Job


class Xqute:
    """The main class of the package

    Attributes:
        name: The name, used in logger
        EMPTY_BUFFER_SLEEP_TIME: The time to sleep while waiting when
            the buffer is empty to wait for the jobs to be pushed

        jobs: The jobs registry
        plugins: The plugins to be enabled or disabled
            to disable a plugin, using `-plugin_name`
            either all plugin names should be prefixed with '+'/'-' or none
            of them should

        _cancelling: A mark to mark whether a shutting down event
            is triggered (True for natural cancelling, the signal for
            cancelling with a signal, SIGINT for example)

        buffer_queue: A buffer queue to save the pushed jobs
        queue: The job queue
        scheduler: The scheduler
        task: The task of producer and consumers

    Args:
        scheduler: The scheduler class or name
        plugins: The plugins to be enabled or disabled
            to disable a plugin, using `-plugin_name`
            either all plugin names should be prefixed with '+'/'-' or none
            of them should
        workdir: The job meta directory
        submission_batch: The number of consumers to submit jobs. This allows
            multiple jobs to be submitted in parallel. This is useful when
            there are many jobs to be submitted and the scheduler has a high
            latency for each submission. Set this to a smaller number if the
            scheduler cannot handle too many simultaneous submissions.
        error_strategy: The strategy when there is error happened
        num_retries: Max number of retries when error_strategy is retry
        forks: Max number of job forks for scheduler
        scheduler_opts: Additional keyword arguments for scheduler
    """

    name: str = "Xqute"
    EMPTY_BUFFER_SLEEP_TIME: int = 1

    def __init__(
        self,
        scheduler: str | Type[Scheduler] = "local",
        *,
        plugins: List[Any] | None = None,
        workdir: str | PathType = DEFAULT_WORKDIR,
        submission_batch: int | None = None,
        error_strategy: str = DEFAULT_ERROR_STRATEGY,
        num_retries: int = DEFAULT_NUM_RETRIES,
        forks: int = DEFAULT_SCHEDULER_FORKS,
        scheduler_opts: Mapping[str, Any] | None = None,
        jobname_prefix: str | None = None,
    ) -> None:
        self.jobs: List[Job] = []

        self.plugin_context = plugin.plugins_context(plugins)

        self.plugin_context.__enter__()

        logger.info(
            "/%s Enabled plugins: %s",
            self.name,
            plugin.get_enabled_plugin_names(),
        )

        self._cancelling: bool | signal.Signals = False
        self._keep_feeding: bool = False
        self._completion_task: asyncio.Task | None = None

        self.buffer_queue: deque = deque()
        self._buffer_event: asyncio.Event = asyncio.Event()
        self.queue: asyncio.Queue = asyncio.Queue()

        scheduler_opts = scheduler_opts or {}
        self.scheduler = get_scheduler(scheduler)(
            workdir=workdir,
            forks=forks,
            error_strategy=error_strategy,
            num_retries=num_retries,
            jobname_prefix=jobname_prefix,
            submission_batch=submission_batch,
            **scheduler_opts,
        )

        # requires to be defined in a loop
        loop = asyncio.get_running_loop()
        for sig in (signal.SIGTERM, signal.SIGINT):
            loop.add_signal_handler(sig, functools.partial(self.cancel, sig))

        self._tasks: asyncio.Task | None = None

        logger.debug("/%s Calling on_init hook ...", self.name)
        plugin.hooks.on_init(self)

    def __del__(self) -> None:
        """Destructor to warn if stop_feeding was not called"""
        if hasattr(self, "is_feeding") and self.is_feeding():
            logger.warning(
                "/%s Instance destroyed while still in keep_feeding mode. "
                "Did you forget to call 'await xqute.stop_feeding()'?",
                self.name,
            )

    def cancel(self, sig: signal.Signals | None = None) -> None:
        """Cancel the producer-consumer task

        `self._cancelling` will be set to `signaled` if sig is provided,
        otherwise it will be set to `True`

        Args:
            sig: Whether this cancelling is caused by a signal
        """
        self._cancelling = True
        if sig:
            self._cancelling = sig
            logger.warning(
                "/%s Got signal %r, trying a graceful " "shutdown ...",
                self.name,
                sig.name,
            )

        logger.debug("/%s Calling on_shutdown hook ...", self.name)

        # Always cancel tasks if not already cancelled, regardless of hook result
        # This prevents plugins from accidentally leaving tasks running
        if plugin.hooks.on_shutdown(self, sig) is not False:
            if self._tasks:
                self._tasks.cancel()

    async def _producer(self) -> None:
        """The producer"""
        polling_counter = 0

        try:
            while True:
                if not self.buffer_queue:
                    # If not in keep_feeding mode and buffer is empty, exit
                    if not self._keep_feeding:
                        logger.debug(
                            "/Producer Buffer empty and not in keep_feeding mode, "
                            "exiting ..."
                        )
                        break
                    logger.debug("/Producer Buffer queue is empty, waiting ...")
                    # Wait for buffer event instead of sleep polling
                    await self._buffer_event.wait()
                    self._buffer_event.clear()
                    continue

                job = self.buffer_queue.popleft()
                # Lightweight check: just count running jobs, no hooks
                n_running = await self.scheduler.count_running_jobs(self.jobs)
                if n_running >= self.scheduler.forks:
                    logger.debug("/Producer Hit max forks of scheduler ...")
                    self.buffer_queue.appendleft(job)
                    # Wait longer when hitting max forks to reduce polling overhead
                    await asyncio.sleep(SLEEP_INTERVAL_PRODUCER_MAX_FORKS)
                    polling_counter += 1
                    continue

                await job.set_status(JobStatus.QUEUED)
                await self.queue.put(job)
                polling_counter = 0  # Reset counter after successful queuing

            # Send sentinel values to stop consumers
            logger.debug("/Producer Finished, sending sentinels to consumers ...")
            for _ in range(self.scheduler.subm_batch):
                await self.queue.put(None)
        except asyncio.CancelledError:
            logger.debug("/Producer Cancelled ...")

    async def _consumer(self, index: int) -> None:
        """The consumer

        Args:
            index: The index of the consumer
        """
        try:
            while True:
                job = await self.queue.get()
                # Check for sentinel value to exit gracefully
                if job is None:
                    logger.debug("/Consumer-%s Received sentinel, exiting ...", index)
                    self.queue.task_done()
                    break

                logger.debug("/Consumer-%s submitting %s", index, job)
                await self.scheduler.submit_job_and_update_status(job)
                self.queue.task_done()
        except asyncio.CancelledError:
            logger.warning("/Consumer-%s Cancelled while submitting ...", index)

    async def feed(self, cmd: CommandType | Job, envs: dict[str, Any] = None) -> None:
        """Put a command into the buffer

        Args:
            cmd: The command
            envs: The environment variables for the job
        """
        from .job import Job

        envs = envs or {}

        if isinstance(cmd, Job):
            job = cmd
            job.envs.update(envs)
        else:
            job = await self.scheduler.create_job(len(self.jobs), cmd, envs)

        logger.debug("/Job-%s Calling on_job_init hook ...", job.index)
        await plugin.hooks.on_job_init(self.scheduler, job)
        self.jobs.append(job)
        logger.info("/%s Pushing job: %r", self.name, job)

        self.buffer_queue.append(job)
        # Signal producer that buffer has new jobs
        self._buffer_event.set()
        logger.debug("/Job-%s Calling on_job_queued hook ...", job.index)
        await plugin.hooks.on_job_queued(self.scheduler, job)

    def is_feeding(self) -> bool:
        """Check if the system is in keep_feeding mode.

        Returns:
            True if in keep_feeding mode and waiting for stop_feeding() to be called.
        """
        return (
            hasattr(self, "_keep_feeding")
            and self._keep_feeding
            and hasattr(self, "_completion_task")
            and self._completion_task
        )

    async def stop_feeding(self) -> None:
        """Stop feeding mode and wait for all jobs to complete.

        After calling this method, the producer will exit once the buffer
        queue is empty, and this method will wait for all jobs to complete.
        This should be called after all jobs have been submitted when using
        run_until_complete(keep_feeding=True).

        Raises:
            RuntimeError: If called without first calling
                run_until_complete(keep_feeding=True)
        """
        if not self.is_feeding():
            logger.error(
                "/%s stop_feeding() called but keep_feeding mode was not started. "
                "Ignoring ...",
                self.name,
            )
            return

        logger.debug("/%s Stopping feeding mode", self.name)
        self._keep_feeding = False

        # Wait for completion if we started in keep_feeding mode
        try:
            await self._completion_task
        except asyncio.CancelledError:  # pragma: no cover
            if self._tasks:
                self._tasks.cancel()

        self._completion_task = None

    async def _polling_jobs(self) -> None:
        """Polling the jobs to see if they are all done.

        If yes, cancel the producer-consumer task naturally.
        """
        try:
            # Wait for feeding to stop if in keep_feeding mode
            while self._keep_feeding:
                await asyncio.sleep(SLEEP_INTERVAL_KEEP_FEEDING)

            polling_counter = 0
            while self._cancelling is False and not await self.scheduler.check_all_done(
                self.jobs, polling_counter
            ):
                await asyncio.sleep(SLEEP_INTERVAL_POLLING_JOBS)
                polling_counter += 1

            if self._cancelling is False:
                self.cancel()
        except asyncio.CancelledError:
            logger.debug("/%s Polling cancelled ...", self.name)
            if self._cancelling not in (True, False):  # signaled
                await self.scheduler.kill_running_jobs(self.jobs)

    async def run_until_complete(self, keep_feeding: bool = False) -> None:
        """Wait until all jobs complete

        Args:
            keep_feeding: If True, starts running in background and returns immediately,
                allowing jobs to be added after calling this method.
                You must call stop_feeding() when done adding jobs, which will
                wait for all jobs to complete.
                If False (default), waits for all current jobs to complete immediately.

        Examples:
            Traditional usage:
            ```python
            xqute = Xqute()
            await xqute.feed(['echo', '1'])
            await xqute.feed(['echo', '2'])
            await xqute.run_until_complete()
            ```

            Keep feeding mode:
            ```python
            xqute = Xqute()
            await xqute.feed(['echo', '1'])
            await xqute.run_until_complete(keep_feeding=True)  # Returns immediately
            await xqute.feed(['echo', '2'])  # Can add more jobs
            await xqute.stop_feeding()  # Waits for completion
            ```
        """
        self._keep_feeding = keep_feeding

        if keep_feeding:
            # Start completion tasks in background
            logger.debug("/%s Starting in keep_feeding mode ...", self.name)
            self._completion_task = asyncio.create_task(self._run_completion_tasks())
            # Return immediately to allow more jobs to be added
            return

        # Traditional mode - wait for completion
        logger.debug(
            "/%s Done feeding jobs, waiting for jobs to be done ...", self.name
        )
        await self._run_completion_tasks()

    async def _run_completion_tasks(self) -> None:
        """Run the completion tasks (polling and await)"""
        self._tasks = asyncio.gather(
            self._producer(),
            *(self._consumer(i) for i in range(self.scheduler.subm_batch)),
            self._polling_jobs(),
        )
        try:
            await self._tasks
        except asyncio.CancelledError:
            logger.debug("/%s Completion tasks cancelled ...", self.name)
        finally:
            logger.info("/%s Done!", self.name)
            if self.plugin_context:
                self.plugin_context.__exit__()



"""The scheduler to run jobs via containers"""

from __future__ import annotations

import asyncio
import os
import shlex
import shutil
from pathlib import Path
from typing import List, Sequence

from ..job import Job
from ..path import SpecPath
from ..defaults import JOBCMD_WRAPPER_LANG
from .local_scheduler import LocalScheduler
from .gbatch_scheduler import NAMED_MOUNT_RE, DEFAULT_MOUNTED_ROOT

CONTAINER_TYPES = {
    "docker": "docker",
    "podman": "podman",
    "apptainer": "apptainer",
    "singularity": "apptainer",
}


class ContainerScheduler(LocalScheduler):
    """Scheduler to run jobs via containers (Docker/Podman/Apptainer)

    This scheduler can execute jobs inside containers using Docker, Podman,
    or Apptainer.

    Args:
        image: Container image to use for running jobs
        entrypoint: Entrypoint command for the container
        bin: Path to container runtime binary (e.g. /path/to/docker)
        volumes: host:container volume mapping string or strings
            or named volume mapping like `MOUNTED=/path/on/host`
            then it will be mounted to `/mnt/disks/MOUNTED` in the container.
            You can use environment variable `MOUNTED` in your job scripts to
            refer to the mounted path.
        user: User to run the container as (only for Docker/Podman)
            By default, it runs as the current user (os.getuid() and os.getgid())
        remove: Whether to remove the container after execution.
            Only applies to Docker/Podman.
        bin_args: Additional arguments to pass to the container runtime
        **kwargs: Additional arguments passed to parent Scheduler
    """

    name = "container"

    __slots__ = (
        "image",
        "entrypoint",
        "bin",
        "volumes",
        # "envs",
        "remove",
        "user",
        "bin_args",
        "_container_type",
        "_path_envs",
    )

    def __init__(
        self,
        image: str,
        entrypoint: str | List[str] = JOBCMD_WRAPPER_LANG,
        bin: str = "docker",
        volumes: str | Sequence[str] | None = None,
        # envs: Dict[str, str] | None = None,
        remove: bool = True,
        user: str | None = None,
        bin_args: List[str] | None = None,
        **kwargs,
    ):
        if "mount" in kwargs:
            raise ValueError(
                "You used 'mount' argument for container scheduler, "
                "did you mean 'volumes'?"
            )

        kwargs.setdefault("mounted_workdir", f"{DEFAULT_MOUNTED_ROOT}/xqute_workdir")
        super().__init__(**kwargs)

        self.bin = shutil.which(bin)
        if not self.bin:
            raise ValueError(f"Container runtime binary '{bin}' not found in PATH")

        self.image = image
        self.entrypoint = (
            list(entrypoint) if isinstance(entrypoint, (list, tuple)) else [entrypoint]
        )
        self._path_envs = {}
        self.volumes = volumes or []
        self.volumes = (
            [self.volumes] if isinstance(self.volumes, str) else list(self.volumes)
        )
        for i, vol in enumerate(self.volumes):
            if NAMED_MOUNT_RE.match(vol):
                name, host_path = vol.split("=", 1)
                host_path_obj = Path(host_path).expanduser().resolve()
                if not host_path_obj.exists():
                    raise FileNotFoundError(
                        f"Volume host path '{host_path}' does not exist"
                    )
                if host_path_obj.is_file():
                    host_path = str(host_path_obj.parent)
                    mount_path = (
                        f"{DEFAULT_MOUNTED_ROOT}/{name}/{host_path_obj.parent.name}"
                    )
                    self._path_envs[name] = f"{mount_path}/{host_path_obj.name}"
                    self.volumes[i] = f"{host_path}:{mount_path}"
                else:
                    host_path = str(host_path_obj)
                    mount_path = f"{DEFAULT_MOUNTED_ROOT}/{name}"
                    self._path_envs[name] = mount_path
                    self.volumes[i] = f"{host_path}:{mount_path}"

        # self.envs = envs or {}
        self.remove = remove
        self.user = user or f"{os.getuid()}:{os.getgid()}"
        self.bin_args = bin_args or []
        self.volumes.append(f"{self.workdir}:{self.workdir.mounted}")

        self._container_type = CONTAINER_TYPES.get(
            Path(self.bin).name.lower(),
            "docker",
        )
        if self._container_type in ("docker", "podman") and self.image.startswith(
            "docker://"
        ):
            # Convert docker://image to image name
            self.image = self.image[9:]

    async def wrapped_job_script(self, job: Job) -> SpecPath:
        """Get the wrapped job script

        Args:
            job: The job

        Returns:
            The path of the wrapped job script
        """
        base = f"job.wrapped.{self.name}-{self._container_type}"
        wrapt_script = job.metadir / base
        await wrapt_script.a_write_text(self.wrap_job_script(job))

        return wrapt_script

    def jobcmd_shebang(self, job: Job) -> str:
        """The shebang of the wrapper script"""
        cmd = [self.bin, "run"]
        if self._container_type == "apptainer":
            if self.cwd:  # pragma: no cover
                cmd.extend(["--pwd", self.cwd])
            else:
                cmd.extend(["--pwd", str(self.workdir.mounted)])
            for key, value in job.envs.items():
                cmd.extend(["--env", f"{key}={value}"])
            for vol in self.volumes:
                cmd.extend(["--bind", f"{vol}"])
        else:
            if self.remove:
                cmd.append("--rm")
            cmd.extend(["--user", self.user])
            for key, value in job.envs.items():
                cmd.extend(["-e", f"{key}={value}"])
            for vol in self.volumes:
                cmd.extend(["-v", vol])

            if self.cwd:
                cmd.extend(["--workdir", self.cwd])
            else:
                cmd.extend(["--workdir", str(self.workdir.mounted)])

        cmd.extend(self.bin_args)
        cmd.append(self.image)
        cmd.extend(self.entrypoint)

        return shlex.join(cmd)

    async def submit_job(self, job: Job) -> int:  # type: ignore[override]
        """Submit a job locally

        Args:
            job: The job

        Returns:
            The process id
        """
        wrapt_script_path = (await self.wrapped_job_script(job)).mounted
        # In case the process exits very quickly
        if not await job.jid_file.a_exists():
            await job.jid_file.a_write_text("0")

        proc = await asyncio.create_subprocess_exec(
            *shlex.split(self.jobcmd_shebang(job)),
            wrapt_script_path,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
            start_new_session=True,
            # Changing the working directory here may cause wrapped_job_script to fail
            # to be found, so we don't set cwd here.
            # The cwd is changed in the wrapper script instead.
            # cwd=self.cwd
        )

        # wait for a while to make sure the process is running
        # this is to avoid the real command is not run when proc is recycled too early
        # this happens for python < 3.12
        await asyncio.sleep(0.1)

        if await job.stdout_file.a_exists():  # pragma: no cover
            # job submitted successfully and already started very soon
            return proc.pid

        if proc.returncode is not None and proc.returncode != 0:
            # The process has already finished and no stdout/stderr files are
            # generated
            # Something went wrong with the wrapper script?
            stderr = await proc.stdout.read()
            raise RuntimeError(
                f"Failed to submit job #{job.index} (rc={proc.returncode}): "
                f"{stderr.decode()}\n"
                f"Command: {self.jobcmd_shebang(job)} "
                f"{wrapt_script_path}\n"
            )

        # don't await for the results, as this will run the real command
        return proc.pid

    def jobcmd_init(self, job) -> str:
        init_cmd = super().jobcmd_init(job)
        path_envs_exports = [
            f"export {key}={shlex.quote(value)}"
            for key, value in self._path_envs.items()
        ]
        if path_envs_exports:
            path_envs_exports.insert(0, "# Mounted paths")
            init_cmd = "\n".join(path_envs_exports) + "\n" + init_cmd

        return init_cmd



"""Utilities for xqute"""

from __future__ import annotations

import os
import logging
from typing import Union, Tuple, List

from rich.logging import RichHandler

from .defaults import LOGGER_NAME, LOGGER_LEVEL

CommandType = Union[str, Tuple[str], List[str]]


logger = logging.getLogger(LOGGER_NAME)
logger.addHandler(RichHandler(show_path=False, omit_repeated_times=False))

loglevel = os.getenv("XQUTE_LOG_LEVEL", LOGGER_LEVEL).upper()
logger.setLevel(loglevel)



"""The scheduler to schedule jobs"""

from __future__ import annotations

import os
import shlex
import signal
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, List, Type

from panpath import CloudPath
from diot import Diot  # type: ignore

from .defaults import (
    JobStatus,
    JobErrorStrategy,
    JOBCMD_WRAPPER_LANG,
    JOBCMD_WRAPPER_TEMPLATE,
    DEFAULT_ERROR_STRATEGY,
    DEFAULT_NUM_RETRIES,
    DEFAULT_SUBMISSION_BATCH,
    DEFAULT_RECHECK_INTERVAL,
    get_jobcmd_wrapper_init,
)
from .utils import logger, CommandType
from .path import SpecPath
from .job import Job
from .plugin import plugin


class Scheduler(ABC):
    """The abstract class for scheduler

    Attributes:
        name: The name of the scheduler
        jobcmd_wrapper_init: The init script for the job command wrapper

    Args:
        workdir: The working directory
        forks: Max number of job forks
        error_strategy: The strategy when there is error happened
        num_retries: Max number of retries when error_strategy is retry
        prescript: The prescript to run before the job command
            It is a piece of script that inserted into the wrapper script, running
            on the scheduler system.
        postscript: The postscript to run when job finished
            It is a piece of script that inserted into the wrapper script, running
            on the scheduler system.
        jobname_prefix: The prefix for the job name
        submission_batch: The number of consumers to submit jobs. This allows
            multiple jobs to be submitted in parallel. This is useful when
            there are many jobs to be submitted and the scheduler has a high
            latency for each submission. Set this to a smaller number if the
            scheduler cannot handle too many simultaneous submissions.
        recheck_interval: The number of polling iterations between rechecks of
            whether a job is still running on the scheduler. Helps detect jobs
            that fail before the wrapped script updates status (e.g., resource
            allocation failures). Each iteration takes
            `xqute.defaults.SLEEP_INTERVAL_POLLING_JOBS`
        cwd: The working directory for the job command wrapper
        **kwargs: Other arguments for the scheduler
    """

    __slots__ = (
        "config",
        "forks",
        "workdir",
        "error_strategy",
        "num_retries",
        "prescript",
        "postscript",
        "jobname_prefix",
        "recheck_interval",
        "subm_batch",
        "cwd",
    )

    # The name of the scheduler
    name: str
    # The number of consumers to submit jobs in parallel
    submission_batch: int = DEFAULT_SUBMISSION_BATCH
    job_class: Type[Job] = Job

    def __init__(
        self,
        workdir: str | Path,
        forks: int = 1,
        error_strategy: str = DEFAULT_ERROR_STRATEGY,
        num_retries: int = DEFAULT_NUM_RETRIES,
        prescript: str = "",
        postscript: str = "",
        jobname_prefix: str | None = None,
        submission_batch: int | None = None,
        recheck_interval: int = DEFAULT_RECHECK_INTERVAL,
        cwd: str | Path = None,
        **kwargs,
    ):
        self.forks = forks
        mounted_workdir = kwargs.pop("mounted_workdir", None)
        self.workdir = SpecPath(workdir, mounted=mounted_workdir)

        self.error_strategy = error_strategy
        self.num_retries = num_retries
        self.prescript = prescript
        self.postscript = postscript
        self.jobname_prefix = jobname_prefix or self.name
        self.subm_batch = submission_batch or self.__class__.submission_batch
        self.recheck_interval = recheck_interval
        self.cwd = None if cwd is None else str(cwd)

        self.config = Diot(**kwargs)

    async def create_job(
        self,
        index: int,
        cmd: CommandType,
        envs: dict[str, Any] | None = None,
    ) -> Job:
        """Create a job

        Args:
            index: The index of the job
            cmd: The command of the job

        Returns:
            The job
        """
        job = self.job_class(
            index=index,
            cmd=cmd,
            workdir=self.workdir,
            error_retry=self.error_strategy == JobErrorStrategy.RETRY,
            num_retries=self.num_retries,
            envs=envs,
        )
        logger.debug("/Job-%s Creating metadir: %s", job.index, job.metadir)
        await job.metadir.a_mkdir(parents=True, exist_ok=True)
        return job

    async def submit_job_and_update_status(self, job: Job):
        """Submit and update the status

        1. Check if the job is already submitted or running
        2. If not, run the hook
        3. If the hook is not cancelled, clean the job
        4. Submit the job, raising an exception if it fails
        5. If the job is submitted successfully, update the status
        6. If the job fails to submit, update the status and write stderr to
            the job file

        Args:
            job: The job
        """
        if await self.job_is_submitted_or_running(job):
            logger.warning(
                "/Job-%s Skip submitting, already submitted or running.",
                job.index,
            )
            return

        exception: Exception | None = None
        try:
            logger.debug("/Job-%s Calling on_job_submitting hook ...", job.index)
            if await plugin.hooks.on_job_submitting(self, job) is False:
                logger.info("/Job-%s submission cancelled by hook.", job.index)
                return

            logger.debug("/Job-%s Cleaning up before submission", job.index)
            await job.clean()

            try:
                # raise the exception immediately
                # it somehow cannot be catched immediately
                logger.debug(
                    "/Job-%s Submitting to scheduler '%s' ...",
                    job.index,
                    self.name,
                )
                await job.set_jid(await self.submit_job(job))
            except Exception as exc:
                exception = RuntimeError(f"Failed to submit job: {exc}")
                exception.__traceback__ = exc.__traceback__
            else:
                await self.transition_job_status(job, JobStatus.SUBMITTED)

        except Exception as exc:  # pragma: no cover
            exception = exc

        if exception is not None:
            from traceback import format_exception

            error_msg = "".join(
                format_exception(
                    type(exception),
                    exception,
                    exception.__traceback__,
                )
            )
            await job.stderr_file.a_write_text(error_msg)
            await self.transition_job_status(job, JobStatus.FAILED, rc="-2")

    async def transition_job_status(
        self,
        job: Job,
        new_status: int,
        old_status: int | None = None,
        flush: bool = True,
        rc: str | None = None,
        error_msg: str | None = None,
        is_killed: bool = False,
    ):
        """Centralized status transition handler

        Handles all aspects of job status transitions:
        - Status change logging
        - Hook lifecycle management (ensuring on_job_started is called)
        - Appropriate hook calls based on new status
        - RC file updates
        - Error message appending to stderr
        - JID file cleanup for terminal states
        - Pipeline halt on errors if configured

        Note that this method will not flush status changes to disk (job.status_file).
        You need to call job.set_status() separately if needed.

        Args:
            job: The job to transition
            new_status: The new status to transition to
            old_status: The previous status (if known).
                If None, will use job._status
            flush: Whether to flush the status to disk
            rc: Optional return code to write to rc_file
            error_msg: Optional error message to append to stderr_file
            is_killed: Whether this is a killed job (uses on_job_killed hook)
        """
        # Save the previous status before updating
        # (setter will update prev_status, so we need to save it first)
        old_status = job._status if old_status is None else old_status
        await job.set_status(new_status, flush=flush)

        # Handle killed jobs specially
        if is_killed:
            logger.debug("/Job-%s Calling on_job_killed hook ...", job.index)
            await plugin.hooks.on_job_killed(self, job)
            return

        # Handle status-specific logic
        if new_status == JobStatus.FAILED:
            # Ensure lifecycle hook was called
            # Job may go from SUBMITTED to FAILED directly (finished too soon)
            if old_status != JobStatus.RUNNING:
                await self.transition_job_status(
                    job,
                    JobStatus.RUNNING,
                    old_status=old_status,
                    flush=False,
                )
                await self.transition_job_status(
                    job,
                    new_status,
                    old_status=JobStatus.RUNNING,
                    flush=False,
                )
                return

            # Write rc file if provided
            if rc is not None:  # pragma: no cover
                await job.set_rc(rc)

            # Append error message if provided
            if error_msg is not None:  # pragma: no cover
                async with job.stderr_file.a_open("a") as f:
                    f.write(f"\n{error_msg}\n")

            # Call failure hook
            logger.debug("/Job-%s Calling on_job_failed hook ...", job.index)
            await plugin.hooks.on_job_failed(self, job)

            # Clean up jid file
            await job.jid_file.a_unlink(missing_ok=True)

            # Handle halt strategy
            if self.error_strategy == JobErrorStrategy.HALT:
                logger.error(
                    "/Sched-%s Pipeline will halt since job failed: %r",
                    self.name,
                    job,
                )
                os.kill(os.getpid(), signal.SIGTERM)

        elif new_status == JobStatus.FINISHED:
            # Ensure lifecycle hook was called
            if old_status != JobStatus.RUNNING:  # pragma: no cover
                await self.transition_job_status(
                    job,
                    JobStatus.RUNNING,
                    old_status=old_status,
                    flush=False,
                )
                await self.transition_job_status(
                    job,
                    new_status,
                    old_status=JobStatus.RUNNING,
                    flush=False,
                )
                return

            # Call success hook
            logger.debug("/Job-%s Calling on_job_succeeded hook ...", job.index)
            await plugin.hooks.on_job_succeeded(self, job)

            # Clean up jid file
            await job.jid_file.a_unlink(missing_ok=True)

        elif new_status == JobStatus.RUNNING:
            # Call started hook
            logger.debug("/Job-%s Calling on_job_started hook ...", job.index)
            await plugin.hooks.on_job_started(self, job)

        elif new_status == JobStatus.SUBMITTED:
            # Call submitted hook
            logger.info(
                "/Sched-%s Job %s submitted (jid: %s, wrapped: %s)",
                self.name,
                job.index,
                await job.get_jid(),
                await self.wrapped_job_script(job),
            )
            logger.debug("/Job-%s Calling on_job_submitted hook ...", job.index)
            await plugin.hooks.on_job_submitted(self, job)

    async def kill_job_and_update_status(self, job: Job):
        """Kill a job and update its status

        Args:
            job: The job
        """
        await job.set_status(JobStatus.KILLING)
        logger.debug("/Job-%s Calling on_job_killing hook ...", job.index)
        ret = await plugin.hooks.on_job_killing(self, job)
        if ret is False:
            logger.info(
                "/Sched-%s Job %s killing cancelled by hook.",
                self.name,
                job.index,
            )
            return

        logger.warning("/Sched-%s Killing job %s ...", self.name, job.index)
        await self.kill_job(job)

        await self.transition_job_status(job, JobStatus.FINISHED, is_killed=True)

    async def retry_job(self, job: Job):
        """Retry a job

        Args:
            job: The job
        """
        await job.set_jid("")
        await job.clean(retry=True)
        job.trial_count += 1
        logger.warning(
            "/Sched-%s Retrying (#%s) job: %r",
            self.name,
            job.trial_count,
            job,
        )
        await self.transition_job_status(job, JobStatus.QUEUED, flush=False)
        await plugin.hooks.on_job_queued(self, job)
        await self.submit_job_and_update_status(job)

    async def count_running_jobs(self, jobs: List[Job]) -> int:
        """Count currently running/active jobs (lightweight check)

        This is optimized for the producer to check if new jobs can be submitted.
        It only counts jobs without refreshing status or calling hooks.

        Args:
            jobs: The list of jobs

        Returns:
            Number of jobs currently in active states
        """
        # check_job_done() has updated _status already
        # statuses = await asyncio.gather(*(job.get_status() for job in jobs))
        return sum(
            1
            for job in jobs
            if job._status
            in (
                JobStatus.QUEUED,
                JobStatus.SUBMITTED,
                JobStatus.RUNNING,
                JobStatus.KILLING,
            )
        )

    async def _check_job_done(self, job: Job, polling_counter: int) -> bool | str:
        """Check if a single job is done (lightweight check)

        This is optimized for the producer to check if new jobs can be submitted.
        It only checks the status without refreshing or calling hooks.

        Args:
            job: The job
            polling_counter: The polling counter for hook calls

        Returns:
            True if the job is done, False if not.
            If the job failed, return the "failed".
        """
        prev_status = job._status
        status = await job.get_status(refresh=True)
        # Keep the previous status, which will be updated later
        # So that later parts can know the previous status
        job._status = prev_status

        # Status changed
        if status != prev_status:
            await self.transition_job_status(
                job,
                status,
                old_status=prev_status,
                flush=False,
            )

        elif status == JobStatus.SUBMITTED:
            # Status keep being SUBMITTED
            # Check if the job fails before running
            if await self.job_fails_before_running(job):  # pragma: no cover
                logger.warning(
                    "/Sched-%s Job %s seems to fail before running, "
                    "check your scheduler logs if necessary.",
                    self.name,
                    job.index,
                )
                await self.transition_job_status(
                    job,
                    JobStatus.FAILED,
                    old_status=prev_status,
                    flush=False,
                    rc="-3",
                    error_msg=(
                        "Error: job seems to fail before running.\n"
                        "Check your scheduler logs if necessary."
                    ),
                )
                # transition_job_status handles the HALT, but we still need to break
                if self.error_strategy == JobErrorStrategy.HALT:
                    return "failed"

        elif status == JobStatus.RUNNING:
            # Status keep being RUNNING
            logger.debug(
                "/Sched-%s Job %s is running, calling polling hook ...",
                self.name,
                job.index,
            )
            # Call the polling hook
            logger.debug("/Job-%s Calling on_job_polling hook ...", job.index)
            await plugin.hooks.on_job_polling(self, job, polling_counter)
            # Let's make sure the job is really running
            if (
                not await job.rc_file.a_is_file()
                and (polling_counter + 1) % self.recheck_interval == 0
                and not await self.job_is_running(job)
            ):  # pragma: no cover
                logger.warning(
                    "/Sched-%s Job %s is not running in the scheduler, "
                    "but its status is still RUNNING, setting it to FAILED",
                    self.name,
                    job.index,
                )
                await self.transition_job_status(
                    job,
                    JobStatus.FAILED,
                    old_status=prev_status,
                    flush=False,
                    rc="-3",
                    error_msg=(
                        "Error: job is not running in the scheduler, "
                        "but its status is still RUNNING.\n"
                        "It is likely that the resource is preempted."
                    ),
                )
                # transition_job_status handles the HALT, but we still need to break
                if self.error_strategy == JobErrorStrategy.HALT:
                    return "failed"

        # Check if we need to halt - this catches any FAILED status
        # transition_job_status already sent SIGTERM, we just need to break the loop
        if self.error_strategy == JobErrorStrategy.HALT and status == JobStatus.FAILED:
            return "failed"

        if status not in (JobStatus.FINISHED, JobStatus.FAILED):
            logger.debug(
                "/Sched-%s Not all jobs are done yet, job %s is %s",
                self.name,
                job.index,
                JobStatus.get_name(status),
            )
            return False

        # Try to resubmit the job for retrying
        if (
            status == JobStatus.FAILED
            and job._error_retry
            and job.trial_count < job._num_retries
        ):
            logger.debug(
                "/Sched-%s Job %s is retrying ...",
                self.name,
                job.index,
            )
            await self.retry_job(job)
            return False

        return True

    async def check_all_done(
        self,
        jobs: List[Job],
        polling_counter: int,
    ) -> bool:
        """Check if all jobs are done (full polling with hooks)

        This does complete status refresh and calls all lifecycle hooks.
        Used by the main polling loop to track job completion.

        Args:
            jobs: The list of jobs
            polling_counter: The polling counter for hook calls

        Returns:
            True if all jobs are done, False otherwise
        """
        ret = True
        logger.debug(
            "/Sched-%s Checking if all jobs are done (#%s) ...",
            self.name,
            polling_counter,
        )

        for job in jobs:
            # We don't use gather here, since we may break the loop early
            job_done = await self._check_job_done(job, polling_counter)
            if job_done == "failed":
                ret = False
                break

            if not job_done:
                ret = False

        return ret

    async def kill_running_jobs(self, jobs: List[Job]):
        """Try to kill all running jobs

        Args:
            jobs: The list of jobs
        """
        logger.warning("/Sched-%s Killing running jobs ...", self.name)
        for job in jobs:
            status = await job.get_status()
            if status in (JobStatus.SUBMITTED, JobStatus.RUNNING):
                await self.kill_job_and_update_status(job)

    async def job_is_submitted_or_running(self, job: Job) -> bool:
        """Check if a job is already submitted or running

        Args:
            job: The job

        Returns:
            True if yes otherwise False.
        """
        if await job.jid_file.a_is_file():
            if await self.job_is_running(job):
                await job.set_status(JobStatus.SUBMITTED)
                return True
        return False

    async def job_fails_before_running(self, job: Job) -> bool:
        """Check if a job fails before running.

        For some schedulers, the job might fail before running (after submission).
        For example, the job might fail to allocate resources. In such a case,
        the wrapped script might not be executed, and the job status will not be
        updated (stays in SUBMITTED). We need to check such jobs and mark them as
        FAILED.

        For the instant scheduler, for example, the local scheduler, the failure will
        be immediately reported when submitting the job, so we don't need to check
        such jobs.

        Args:
            job: The job to check

        Returns:
            True if the job fails before running, otherwise False.
        """
        return False

    @property
    def jobcmd_wrapper_init(self) -> str:
        """The init script for the job command wrapper"""
        wrapper_init = get_jobcmd_wrapper_init(
            not isinstance(self.workdir.mounted, CloudPath)
        )
        if self.cwd:
            # Some schedulers (e.g. Google Cloud Batch) doesn't support changing the
            # working directory via configuration, so we need to change it in the
            # wrapper script.
            # See: https://issuetracker.google.com/issues/336164416
            wrapper_init = f"cd {shlex.quote(self.cwd)}\n\n{wrapper_init}"

        return wrapper_init

    def jobcmd_shebang(self, job: Job) -> str:
        """The shebang of the wrapper script"""
        wrapper_lang = (
            JOBCMD_WRAPPER_LANG
            if isinstance(JOBCMD_WRAPPER_LANG, (tuple, list))
            else [JOBCMD_WRAPPER_LANG]
        )
        return shlex.join(wrapper_lang)

    def jobcmd_init(self, job) -> str:
        """The job command init"""
        init_code = []
        if job.envs:
            init_code.append("# Environment variables")
            init_code.extend(
                [
                    f"export {key}={shlex.quote(str(value))}"
                    for key, value in job.envs.items()
                ]
            )

        codes = plugin.hooks.on_jobcmd_init(self, job)
        init_code.extend([code for code in codes if code])
        return "\n".join(init_code)

    def jobcmd_prep(self, job) -> str:
        """The job command preparation"""
        codes = plugin.hooks.on_jobcmd_prep(self, job)
        codes = [code for code in codes if code]
        return "\n".join(codes)

    def jobcmd_end(self, job) -> str:
        """The job command end"""
        codes = plugin.hooks.on_jobcmd_end(self, job)
        codes = [code for code in codes if code]
        return "\n".join(codes)

    def wrap_job_script(self, job: Job) -> str:
        """Wrap the job script

        Args:
            job: The job

        Returns:
            The wrapped script
        """
        return JOBCMD_WRAPPER_TEMPLATE.format(
            scheduler=self,
            shebang=self.jobcmd_shebang(job),
            status=JobStatus,
            job=job,
            cmd=shlex.join(job.cmd),
            jobcmd_init=self.jobcmd_init(job),
            jobcmd_prep=self.jobcmd_prep(job),
            jobcmd_end=self.jobcmd_end(job),
        )

    async def wrapped_job_script(self, job: Job) -> SpecPath:
        """Get the wrapped job script

        Args:
            job: The job

        Returns:
            The path of the wrapped job script
        """
        base = f"job.wrapped.{self.name}"
        wrapt_script = job.metadir / base
        await wrapt_script.a_write_text(self.wrap_job_script(job))

        return wrapt_script

    @abstractmethod
    async def submit_job(self, job: Job) -> int | str:
        """Submit a job

        Args:
            job: The job

        Returns:
            The unique id in the scheduler system
        """

    @abstractmethod
    async def kill_job(self, job: Job):
        """Kill a job

        Args:
            job: The job
        """

    @abstractmethod
    async def job_is_running(self, job: Job) -> bool:
        """Check if a job is really running

        Args:
            job: The job

        Returns:
            True if yes otherwise False.
        """
